/*******************************************************************************
//
// This file is part of the Marble Desktop Globe.
//
// This program is free software licensed under the GNU LGPL. You can
// find a copy of this license in LICENSE.txt in the top directory of
// the source code.
//
// Copyright 2012      Konrad Enzensberger / e.konrad@mpegcode.com
//
********************************************************************************/

/* INCLUDE FILES ***************************************************************/
#include "UTMCoordTrans.h"

#include <QString>
#include <math.h>

UTMCoordTrans::UTMCoordTrans()
{

}
UTMCoordTrans::~UTMCoordTrans()
{

}

/*
' This Function calculates UTM Grid Reference
' from Lat/Long degrees on a Transverse Mercator projection.
' latitude is the latitude in decimal degrees, negative for South.
' longitude is the longitude in decimal degrees, negative for West.*/

bool UTMCoordTrans::LL2UTM(double& latitude,double& longitude,QString& pos)
{
	if(!ValidLat((int)latitude) )
	{
		pos = "#INVALID LATITUDE";
		return false;
	}

	if(!ValidLong((int)longitude) )
	{
		pos = "#INVALID LONGITUDE";
		return false;
	}

	if (latitude < -80 || longitude >= 84)
	{
		pos = "#LATITUDE OUTSIDE MGRS AREA";
		return false;
	}

	latitude =  ( /*(long)*/ (latitude * 1000000 + 1) ) / 1000000;  // round to 0.000001 deg
	longitude = ( /*(long)*/ (longitude * 100000 + 1) ) / 100000;  // round up to 0.00001 deg

	long Merid = UTMCM(latitude,longitude);
	double IEast = LL2TM(true, latitude,longitude, Merid); //Easting
	double INorth = LL2TM(false, latitude,longitude, Merid); //Northing

	if ( INorth < 0) INorth = INorth + 10000000; // S Hemisphere

	long lIEast = (long)(IEast + 0.5);
	long lINorth= (long)(INorth + 0.5);

	IEast = (double)lIEast;
	INorth = (double)lINorth;
	QString EastStr;
	EastStr = QString("%1")
				.arg(abs(lIEast));
	QString NorthStr;
	NorthStr = QString("%1")
				.arg(abs(lINorth));

	QString tmp("0");
	int length = 0;
	do{
		length = EastStr.length();
		if(length<7)
			EastStr = tmp + EastStr;
			
	}while(length<7);

	tmp = "0";
	length = 0;
	do{
		length = NorthStr.length();
		if(length<7)
			NorthStr = tmp + NorthStr;
	
	}while(length<7);

	long LongZone = UTMLongZone(latitude,longitude);
	QString LatZone; 
	UTMLatZone(latitude,LatZone);

	pos = QString("%1%2 %3 %4").arg(LongZone).arg(LatZone).arg(EastStr).arg(NorthStr);

	return true;
}

/*
' This Function calculates a National or UTM (MGRS) Grid Reference
'  from Lat/Long degrees on a Transverse Mercator projection.
' latitude is the latitude in decimal degrees, negative for South.
' longitude is the longitude in decimal degrees, negative for West.*/

bool UTMCoordTrans::LL2GR(double& latitude,double& longitude,QString& pos)
{
	if(!ValidLat((int)latitude) )
	{
		pos = "#INVALID LATITUDE";
		return false;
	}

	if(!ValidLong((int)longitude) )
	{
		pos = "#INVALID LONGITUDE";
		return false;
	}

	if (latitude < -80 || longitude >= 84)
	{
		pos = "#LATITUDE OUTSIDE MGRS AREA";
		return false;
	}

	latitude =  ( /*(long)*/ (latitude * 1000000 + 1) ) / 1000000;  // round to 0.000001 deg
	longitude = ( /*(long)*/ (longitude * 100000 + 1) ) / 100000;  // round up to 0.00001 deg

	long Merid = UTMCM(latitude,longitude);
	double IEast = LL2TM(true, latitude,longitude, Merid); //Easting
	double INorth = LL2TM(false, latitude,longitude, Merid); //Northing

	 if ( INorth < 0) INorth = INorth + 10000000; // S Hemisphere

	long lIEast = (long)(IEast + 0.5);
	long lINorth= (long)(INorth + 0.5);

	IEast = (double)lIEast;
	INorth = (double)lINorth;
	QString EastStr;
	EastStr = QString("%1")
				.arg(abs(lIEast));
	QString NorthStr;
	NorthStr = QString("%1")
				.arg(abs(lINorth));

	QString tmp("0");
	int length = 0;
	do{
		length = EastStr.length();
		if(length<7)
			EastStr = tmp + EastStr;
			
	}while(length<7);

	tmp = "0";
	length = 0;
	do{
		length = NorthStr.length();
		if(length<7)
			NorthStr = tmp + NorthStr;
	
	}while(length<7);

	tmp = EastStr.mid(1, 1) + NorthStr.mid( 1, 1);
	long GR100km = tmp.toLong();;

	QString GRremainder = EastStr.mid(2, 5) + NorthStr.mid( 2, 5);

	long LongZone = UTMLongZone(latitude,longitude);
	QString LatZone;
	UTMLatZone(latitude,LatZone);
	int Pos = (int)(fabs(INorth) / 100000);//Int() ??
	while( Pos > 19)
		Pos = Pos - 20;
	
	if( (LongZone % 2) == 0)
		Pos = Pos + 5;
	if (Pos > 19) Pos = Pos - 20;

/*	if (MGRS_Old)
	{
		Pos = Pos + 10;
		if (Pos > 19) Pos = Pos - 20;
	}*/

	QString Letters = "ABCDEFGHJKLMNPQRSTUVWXYZ";
	QString N100km = Letters.mid(Pos/* + 1*/, 1);//enz pos = +1 ?
	Pos = (int)((GR100km / 10) - 1);//Int()
	long P = LongZone;
	while (P > 3)
		P = P - 3;
	
	Pos = Pos + ((P - 1) * 8);
	QString E100km = Letters.mid(Pos/* + 1*/, 1);//enz pos = +1 ?
	pos = QString("%1%2%3%4%5")
					.arg(LongZone)
					.arg(LatZone)
					.arg(E100km)
					.arg(N100km)
					.arg(GRremainder);
			
	return true;
}

double UTMCoordTrans::LL2TM(bool EastNorth,double& latitude,double& longitude,long Merid)
{
	double ret = 0;

	//parameters
	double F0 = 0;		//Local scale factor on Central Meridian
    double A1 = 0;		//Major Semi-axis, Airy ellipsoid
    double B1 = 0;		//Minor Semi-axis, Airy ellipsoid
    double K0 = 0;		//Lat of True Origin
    double L0 = 0;		//Long of True Origin (2W)
    double N0 = 0;		//Grid Northing of True Origin (m)
    double E0 = 0;		//Grid Easting of True Origin (m)

	if(!ValidLat((int)latitude) )
	{
		ret = -1;//"#INVALID LATITUDE"
		return ret;
	}

	if(!ValidLong((int)longitude) )
	{
		ret = -1;//"#INVALID LONGITUDE"
		return ret;
	}

	if(!ValidLong(Merid) ) 
	{
		ret = -1;//"#INVALID CENTRAL MERIDIAN"
		return ret;
	}

	double Deg2Rad = 3.14159265358979 / 180;

	//UTM WGS84 Ellipsoid
	F0 = 0.9996;
	A1 = 6378137.0 * F0;
	B1 = 6356752.3142 * F0;
	K0 = 0;
	L0 = Merid * Deg2Rad;
	N0 = 0;
	E0 = 500000;

	//Calculation
	double N1 = (A1 - B1) / (A1 + B1); // n
	double N2 = N1 * N1;
	double N3 = N2 * N1;
	double E2 = ((A1 * A1) - (B1 * B1)) / (A1 * A1);// e^2
	double K = latitude * Deg2Rad;
	double L = longitude * Deg2Rad;
	double SINK = sin(K);
	double COSK = cos(K);
	double TANK = SINK / COSK;
	double TANK2 = TANK * TANK;
	double COSK2 = COSK * COSK;
	double COSK3 = COSK2 * COSK;
	double K3 = K - K0;
	double K4 = K + K0;
	double J3 = K3 * (1 + N1 + 1.25 * (N2 + N3));
	double J4 = sin(K3) * cos(K4) * (3 * (N1 + N2 + 0.875 * N3));
	double J5 = sin(2 * K3) * cos(2 * K4) * (1.875 * (N2 + N3));
	double J6 = sin(3 * K3) * cos(3 * K4) * 35 / 24 * N3;
	double M = (J3 - J4 + J5 - J6) * B1;
	double Temp = 1 - E2 * SINK * SINK;
	double V = A1 / sqrt(Temp);
	double R = V * (1 - E2) / Temp;
	double H2 = V / R - 1;
	double P = L - L0;
	double P2 = P * P;
	double P4 = P2 * P2;
	
	J3 = M + N0;
	J4 = V / 2 * SINK * COSK;
	J5 = V / 24 * SINK * (COSK3) * (5 - (TANK2) + 9 * H2);
	J6 = V / 720 * SINK * COSK3 * COSK2 * (61 - 58 * (TANK2) + TANK2 * TANK2);
	double N = J3 + P2 * J4 + P4 * J5 + P4 * P2 * J6;
	double J7 = V * COSK;
	double J8 = V / 6 * COSK3 * (V / R - TANK2);
	double J9 = V / 120 * COSK3 * COSK2;
	J9 = J9 * (5 - 18 * TANK2 + TANK2 * TANK2 + 14 * H2 - 58 * TANK2 * H2);
	double E = E0 + P * J7 + P2 * P * J8 + P4 * P * J9;

	if(EastNorth) //"EASTING"
		ret = E-0.005;
	//'LL2TM = Int(E + 0.5)
	if(!EastNorth) // "NORTHING" 
		ret = N+0.005;
	//'LL2TM = Int(N + 0.5)

	return ret;

}//End Function LL2TM

long UTMCoordTrans::UTMLongZone(double latitude,double longitude)
{
	long ret = 0;
	if(!ValidLat((int)latitude) )
	{
		ret = -1;//"#INVALID LATITUDE"
		return ret;
	}

	if(!ValidLong((int)longitude) )
	{
		ret = -1;//"#INVALID LONGITUDE"
		return ret;
	}

	ret = (UTMCM(latitude,longitude) - 3) / 6 + 31;
	return ret;
}

bool UTMCoordTrans::UTMLatZone(double latitude,QString& LatZone)
{
	QString Letters("ABCDEFGHJKLMNPQRSTUVWXYZ");

	if(!ValidLat((int)latitude) )
	{
		LatZone = "#INVALID LATITUDE";
		return false;
	}

	double dPos = (latitude / 8) + 13;
	if(dPos < 0) dPos += -0.5;
	int Pos = (int) dPos; 

	if(Pos>=(int)Letters.length())
		LatZone = "X";
	else 
	{//Pos - 1 wg. BasiQString Index 1 - enz
		if(Pos>0)Pos--;
		LatZone = Letters.mid(Pos, 1);
	}

//	if (LatZone > "X") Then LatZone = "X")
		
	if( Letters.compare("Y")==0 || Letters.compare("Z")==0)
		LatZone = "X";
	//UTMLatZone = LatZone

	return true;
}

long UTMCoordTrans::UTMCM(double latitude,double longitude)
{
	long ret=0;

	if(!ValidLat((int)latitude) )
	{
	  ret = -1;//"#INVALID LATITUDE"
	  return ret;
	}

	if(!ValidLong((int)longitude) )
	{
	  ret = -1;//"#INVALID LONGITUDE"
	  return ret;
	}

	//Int((LongCell) / 6) * 6 + 3;
	double MeridD = (longitude / 6) ;
	if(MeridD < 0) MeridD--;// += -0.5;
	long Merid = (long)MeridD; 
	Merid = Merid * 6 + 3;

	if( (latitude >= 72) && (longitude >= 0) )
	{
	  if(longitude < 9)
		Merid = 3;
	  else if (longitude < 21)
		Merid = 15;
	  else if (longitude < 33) 
		Merid = 27;
	  else if (longitude < 42)
		Merid = 39;
	}

	if ((latitude >= 56) && (latitude < 64)) 
	  if ((longitude >= 3) && (longitude < 12)) Merid = 9;
	
	ret = Merid;
	return ret;
}

bool UTMCoordTrans::ValidLat(long latitude)
{
	bool ret = false;
	if ( (latitude < -90) || (latitude > 90) )
		ret = false;
	else ret = true;

	return ret;
}

bool UTMCoordTrans::ValidLong(long longitude)
{
	bool ret = false;
	if( (longitude < -180) || (longitude > 180) )
		ret = false;
	else
		ret = true;

	return ret;
}
